@@ -18,24 +18,45 @@
 #include <string.h>
 #include <errno.h>
 #include <stdio.h>
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
+#endif
 #include <stdlib.h>
+#include <ctype.h>
+
+#ifdef G_OS_WIN32
+/* No hardcoded paths in the binary, thanks */
+#undef PKGLIBDIR
+/* It's OK to leak this, as PKGLIBDIR is invoked only once */
+#define PKGLIBDIR g_strconcat (g_win32_get_package_installation_directory (PACKAGE, NULL), "\\lib\\pkgconfig", NULL)
+#endif
 
 static void verify_package (Package *pkg);
 
 static GHashTable *packages = NULL;
 static GHashTable *locations = NULL;
+static GHashTable *path_positions = NULL;
 static GHashTable *globals = NULL;
 static GSList *search_dirs = NULL;
+static int scanned_dir_count = 0;
 
 gboolean disable_uninstalled = FALSE;
 
 void
 add_search_dir (const char *path)
 {
-  search_dirs = g_slist_prepend (search_dirs, g_strdup (path));
+  search_dirs = g_slist_append (search_dirs, g_strdup (path));
 }
 
+#ifdef G_OS_WIN32
+/* Guard against .pc file being installed with UPPER CASE name */
+# define FOLD(x) tolower(x)
+# define FOLDCMP(a, b) g_ascii_strcasecmp (a, b)
+#else
+# define FOLD(x) (x)
+# define FOLDCMP(a, b) strcmp (a, b)
+#endif
+
 #define EXT_LEN 3
 
 static gboolean
@@ -45,8 +66,8 @@ ends_in_dotpc (const char *str)
   
   if (len > EXT_LEN &&
       str[len - 3] == '.' &&
-      str[len - 2] == 'p' &&
-      str[len - 1] == 'c')
+      FOLD (str[len - 2]) == 'p' &&
+      FOLD (str[len - 1]) == 'c')
     return TRUE;
   else
     return FALSE;
@@ -61,7 +82,7 @@ name_ends_in_uninstalled (const char *str)
   int len = strlen (str);
   
   if (len > UNINSTALLED_LEN &&
-      strcmp ((str + len - UNINSTALLED_LEN), "uninstalled") == 0)
+      FOLDCMP ((str + len - UNINSTALLED_LEN), "uninstalled") == 0)
     return TRUE;
   else
     return FALSE;
@@ -74,13 +95,22 @@ name_ends_in_uninstalled (const char *str)
 static void
 scan_dir (const char *dirname)
 {
-  DIR *dir = opendir (dirname);
+  DIR *dir;
   struct dirent *dent;
   int dirnamelen = strlen (dirname);
+  /* Use a copy of dirname cause Win32 opendir doesn't like
+   * superfluous trailing (back)slashes in the directory name.
+   */
+  char *dirname_copy = strdup (dirname);
 
-  if (dirnamelen > 0 && dirname[dirnamelen-1] == '/')
-    dirnamelen--;
+  if (dirnamelen > 1 && dirname[dirnamelen-1] == G_DIR_SEPARATOR)
+    {
+      dirnamelen--;
+      dirname_copy[dirnamelen-1] = '\0';
+    }
 
+  dir = opendir (dirname_copy);
+  free (dirname_copy);
   if (!dir)
     {
       debug_spew ("Cannot open directory '%s' in package search path: %s\n",
@@ -89,6 +119,8 @@ scan_dir (const char *dirname)
     }
 
   debug_spew ("Scanning directory '%s'\n", dirname);
+
+  scanned_dir_count += 1;
   
   while ((dent = readdir (dir)))
     {
@@ -112,11 +144,13 @@ scan_dir (const char *dirname)
             {
               char *filename = g_malloc (dirnamelen + 1 + len + 1);
               strncpy (filename, dirname, dirnamelen);
-              filename[dirnamelen] = '/';
+              filename[dirnamelen] = G_DIR_SEPARATOR;
               strcpy (filename + dirnamelen + 1, dent->d_name);
               
               g_hash_table_insert (locations, pkgname, filename);
-
+              g_hash_table_insert (path_positions, pkgname,
+                                   GINT_TO_POINTER (scanned_dir_count));
+              
               debug_spew ("Will find package '%s' in file '%s'\n",
                           pkgname, filename);
             }
@@ -140,6 +174,7 @@ package_init ()
       
       packages = g_hash_table_new (g_str_hash, g_str_equal);
       locations = g_hash_table_new (g_str_hash, g_str_equal);
+      path_positions = g_hash_table_new (g_str_hash, g_str_equal);
       
       g_slist_foreach (search_dirs, (GFunc)scan_dir, NULL);
       scan_dir (PKGLIBDIR);
@@ -237,6 +272,12 @@ internal_get_package (const char *name, gboolean warn, gboolean check_compat)
       return NULL;
     }
 
+  pkg->path_position =
+    GPOINTER_TO_INT (g_hash_table_lookup (path_positions, pkg->name));
+
+  debug_spew ("Path position of '%s' is %d\n",
+              pkg->name, pkg->path_position);
+  
   if (strstr (location, "uninstalled.pc"))
     pkg->uninstalled = TRUE;
   
@@ -387,15 +428,42 @@ get_requires (Package *pkg)
   return pkg->requires;
 }
 
+static int
+pathposcmp (gconstpointer a, gconstpointer b)
+{
+  const Package *pa = a;
+  const Package *pb = b;
+
+  if (pa->path_position < pb->path_position)
+    return -1;
+  else if (pa->path_position > pb->path_position)
+    return 1;
+  else
+    return 0;
+}
+
+static GSList*
+packages_sort_by_path_position (GSList *list)
+{
+  return g_slist_sort (list, pathposcmp);
+}
+
 static void
-recursive_fill_list (Package *pkg, GetListFunc func, GSList **listp)
+fill_one_level (Package *pkg, GetListFunc func, GSList **listp)
 {
-  GSList *tmp;
   GSList *copy;
 
   copy = g_slist_copy ((*func)(pkg));
 
   *listp = g_slist_concat (*listp, copy);
+}
+
+static void
+recursive_fill_list (Package *pkg, GetListFunc func, GSList **listp)
+{
+  GSList *tmp;
+
+  fill_one_level (pkg, func, listp);
   
   tmp = pkg->requires;
 
@@ -407,6 +475,62 @@ recursive_fill_list (Package *pkg, GetListFunc func, GSList **listp)
     }
 }
 
+static void
+fill_list_in_path_order_single_package (Package *pkg, GetListFunc func,
+                                        GSList **listp)
+{
+  /* First we get the list in natural/recursive order, then
+   * stable sort by path position
+   */
+  GSList *packages;
+  GSList *tmp;
+  
+  packages = g_slist_append (packages, pkg);
+  recursive_fill_list (pkg, get_requires, &packages);
+
+  packages = packages_sort_by_path_position (packages);
+
+  tmp = packages;
+  while (tmp != NULL)
+    {
+      fill_one_level (tmp->data, func, listp);
+      
+      tmp = tmp->next;
+    }
+
+  g_slist_free (packages);
+}
+
+static void
+fill_list_in_path_order (GSList *packages, GetListFunc func,
+                         GSList **listp)
+{
+  GSList *tmp;
+  GSList *expanded;
+
+  expanded = NULL;
+  tmp = packages;
+  while (tmp != NULL)
+    {
+      expanded = g_slist_append (expanded, tmp->data);
+      recursive_fill_list (tmp->data, get_requires, &expanded);
+
+      tmp = tmp->next;
+    }
+
+  expanded = packages_sort_by_path_position (expanded);
+
+  tmp = expanded;
+  while (tmp != NULL)
+    {
+      fill_one_level (tmp->data, func, listp);
+      
+      tmp = tmp->next;
+    }
+
+  g_slist_free (expanded);
+}
+
 static gint
 compare_req_version_names (gconstpointer a, gconstpointer b)
 {
@@ -433,7 +557,8 @@ verify_package (Package *pkg)
   GSList *iter;
   GSList *requires_iter;
   GSList *conflicts_iter;  
-
+  int count;
+  
   /* Be sure we have the required fields */
 
   if (pkg->key == NULL)
@@ -537,6 +662,61 @@ verify_package (Package *pkg)
   
   g_slist_free (requires);
   g_slist_free (conflicts);
+
+  count = 0;
+  iter = pkg->I_cflags;
+  while (iter != NULL)
+    {
+      /* we put things in canonical -I/usr/include (vs. -I /usr/include) format,
+       * but if someone changes it later we may as well be robust
+       */
+      if (strcmp (iter->data, "-I/usr/include") == 0 ||
+          strcmp (iter->data, "-I /usr/include") == 0)
+        {
+          debug_spew ("Package %s has -I/usr/include in Cflags\n",
+                      pkg->name);
+          if (g_getenv ("PKG_CONFIG_ALLOW_SYSTEM_CFLAGS") == NULL)
+            {
+              iter->data = NULL;
+              ++count;
+              debug_spew ("Removing -I/usr/include from cflags for %s\n", pkg->key);              
+            }
+        }
+
+      iter = iter->next;
+    }
+
+  while (count)
+    {
+      pkg->I_cflags = g_slist_remove (pkg->I_cflags, NULL);
+      --count;
+    }
+
+  count = 0;
+  iter = pkg->L_libs;
+  while (iter != NULL)
+    {
+      if (strcmp (iter->data, "-L/usr/lib") == 0 ||
+          strcmp (iter->data, "-L /usr/lib") == 0)
+        {
+          debug_spew ("Package %s has -L/usr/lib in Libs\n",
+                      pkg->name);
+          if (g_getenv ("PKG_CONFIG_ALLOW_SYSTEM_LIBS") == NULL)
+            {              
+              iter->data = NULL;
+              ++count;
+              debug_spew ("Removing -I/usr/lib from libs for %s\n", pkg->key);
+            }
+        }
+
+      iter = iter->next;
+    }
+
+  while (count)
+    {
+      pkg->L_libs = g_slist_remove (pkg->L_libs, NULL);
+      --count;
+    }
 }
 
 static char*
@@ -546,7 +726,7 @@ get_merged (Package *pkg, GetListFunc func)
   GSList *dups_list = NULL;
   char *retval;
   
-  recursive_fill_list (pkg, func, &dups_list);
+  fill_list_in_path_order_single_package (pkg, func, &dups_list);
   
   list = string_list_strip_duplicates (dups_list);
 
@@ -566,7 +746,7 @@ get_merged_from_back (Package *pkg, GetListFunc func)
   GSList *dups_list = NULL;
   char *retval;
   
-  recursive_fill_list (pkg, func, &dups_list);
+  fill_list_in_path_order_single_package (pkg, func, &dups_list);
   
   list = string_list_strip_duplicates_from_back (dups_list);
 
@@ -587,13 +767,7 @@ get_multi_merged (GSList *pkgs, GetListFunc func)
   GSList *list;
   char *retval;
 
-  tmp = pkgs;
-  while (tmp != NULL)
-    {
-      recursive_fill_list (tmp->data, func, &dups_list);  
-      
-      tmp = g_slist_next (tmp);
-    }
+  fill_list_in_path_order (pkgs, func, &dups_list);
   
   list = string_list_strip_duplicates (dups_list);
 
@@ -614,13 +788,7 @@ get_multi_merged_from_back (GSList *pkgs, GetListFunc func)
   GSList *list;
   char *retval;
 
-  tmp = pkgs;
-  while (tmp != NULL)
-    {
-      recursive_fill_list (tmp->data, func, &dups_list);  
-      
-      tmp = g_slist_next (tmp);
-    }
+  fill_list_in_path_order (pkgs, func, &dups_list);
   
   list = string_list_strip_duplicates_from_back (dups_list);
 
@@ -655,7 +823,6 @@ package_get_L_libs (Package *pkg)
     pkg->L_libs_merged = get_merged (pkg, get_L_libs);
 
   return pkg->L_libs_merged;
-
 }
 
 char *
@@ -927,8 +1094,8 @@ static int rpmvercmp(const char * a, const char * b) {
 
     /* loop through each version segment of str1 and str2 and compare them */
     while (*one && *two) {
-	while (*one && !isalnum(*one)) one++;
-	while (*two && !isalnum(*two)) two++;
+	while (*one && !isalnum((guchar)*one)) one++;
+	while (*two && !isalnum((guchar)*two)) two++;
 
 	str1 = one;
 	str2 = two;
@@ -936,13 +1103,13 @@ static int rpmvercmp(const char * a, const char * b) {
 	/* grab first completely alpha or completely numeric segment */
 	/* leave one and two pointing to the start of the alpha or numeric */
 	/* segment and walk str1 and str2 to end of segment */
-	if (isdigit(*str1)) {
-	    while (*str1 && isdigit(*str1)) str1++;
-	    while (*str2 && isdigit(*str2)) str2++;
+	if (isdigit((guchar)*str1)) {
+	    while (*str1 && isdigit((guchar)*str1)) str1++;
+	    while (*str2 && isdigit((guchar)*str2)) str2++;
 	    isnum = 1;
 	} else {
-	    while (*str1 && isalpha(*str1)) str1++;
-	    while (*str2 && isalpha(*str2)) str2++;
+	    while (*str1 && isalpha((guchar)*str1)) str1++;
+	    while (*str2 && isalpha((guchar)*str2)) str2++;
 	    isnum = 0;
 	}
 		
@@ -1085,6 +1252,14 @@ comparison_to_str (ComparisonType comparison)
   return "???";
 }
 
+static void
+max_len_foreach (gpointer key, gpointer value, gpointer data)
+{
+  int *mlen = data;
+
+  *mlen = MAX (*mlen, strlen (key));
+}
+
 static void
 packages_foreach (gpointer key, gpointer value, gpointer data)
 {
@@ -1092,14 +1267,23 @@ packages_foreach (gpointer key, gpointer value, gpointer data)
 
   if (pkg != NULL)
     {
-      printf ("%s \t\t%s - %s\n",
-              pkg->key, pkg->name, pkg->description);
+      char *pad;
+
+      pad = g_strnfill (GPOINTER_TO_INT (data) - strlen (pkg->key), ' ');
+      
+      printf ("%s%s%s - %s\n",
+              pkg->key, pad, pkg->name, pkg->description);
+
+      g_free (pad);
     }
 }
 
 void
 print_package_list (void)
 {
-  g_hash_table_foreach (locations, packages_foreach, NULL);
+  int mlen = 0;
+  
+  g_hash_table_foreach (locations, max_len_foreach, &mlen);
+  g_hash_table_foreach (locations, packages_foreach, GINT_TO_POINTER (mlen + 1));
 }
 
